import uuid
from typing import Literal, Optional

from qdrant_client import models as qdrant_models

from ..schemas import QueryBase

try:
    from qdrant_client.grpc import common_pb2 as grpc_common_pb2  # type: ignore
except Exception:  # pragma: no cover - optional dependency
    grpc_common_pb2 = None


def cleanse_query_for_search(
    query: Optional[QueryBase | dict],
    offset_type: Literal["int", "str"] = "int",
    *,
    allow_order_by: bool = False,
) -> qdrant_models.Filter | None:
    if query is None:
        return None
    elif isinstance(query, dict):
        query = QueryBase.model_validate(query)
    if offset_type == "int":
        assert query.offset is None or isinstance(query.offset, int), (
            "[Db] Offset must be int in Qdrant searching"
        )
    elif offset_type == "str":
        point_id_types: tuple[type, ...] = (int, str, uuid.UUID)
        if grpc_common_pb2 is not None:
            point_id_types = (*point_id_types, grpc_common_pb2.PointId)
        assert query.offset is None or isinstance(query.offset, point_id_types), (
            "[Db] Offset must be point id (int|str|uuid|grpc PointId) in Qdrant scroll"
        )
    assert query.id is None, (
        "[Db] QueryBase's 'id' key should not being use for similarity search in qdrant"
    )
    assert query.ids is None, (
        "[Db] QueryBase's 'ids' key should not being use for similarity search in qdrant"
    )
    if not allow_order_by:
        assert query.order_by is None, (
            "[Db] QueryBase's 'order_by' key should not being use for similarity search in qdrant"
        )
    if query.filters is None:
        return None

    if isinstance(query.filters, qdrant_models.Filter):
        return query.filters

    if not isinstance(query.filters, dict):
        raise AssertionError("[Db] Query.filters only accept dict or qdrant Filter")

    # forbidden id and ids in filters
    if "id" in query.filters:
        raise AssertionError(
            "[Db] You should not include 'id' in your QueryBase.filters when searching Qdrant"
        )
    if "ids" in query.filters:
        raise AssertionError(
            "[Db] You should not include 'ids' in your QueryBase.filters when searching Qdrant"
        )

    return qdrant_models.Filter.model_validate(query.filters)



from typing import AsyncGenerator, Generator, Optional

from motor.motor_asyncio import AsyncIOMotorClient
from neo4j import AsyncSession as Neo4jAsyncSession
from neo4j import AsyncTransaction as Neo4jAsyncTransaction
from neo4j import Session as Neo4jSession
from pymongo import MongoClient
from qdrant_client import AsyncQdrantClient, QdrantClient
from redis import Redis
from redis.asyncio import Redis as AsyncRedis
from sqlalchemy.ext.asyncio import async_sessionmaker
from sqlmodel import Session
from sqlmodel.ext.asyncio.session import AsyncSession as SQLModelAsyncSession

from keble_db.wrapper import ExtendedAsyncRedis
from keble_db.wrapper.redis import ExtendedRedis

from ..session import Db


class ApiDbDeps:
    def __init__(self, db: Db):
        self.__db = db

        # 1) Create your session factory once, ideally at module import, not per-request
        self.self_asql_write_session_maker = (
            async_sessionmaker(
                self.__db.async_sql_write_engine,
                class_=SQLModelAsyncSession,
                expire_on_commit=False,  # keep attributes loaded after commit
            )
            if self.__db.async_sql_write_engine is not None
            else None
        )

        self.self_asql_read_session_maker = (
            async_sessionmaker(
                self.__db.async_sql_read_engine,
                class_=SQLModelAsyncSession,
                expire_on_commit=False,  # keep attributes loaded after commit
            )
            if self.__db.async_sql_read_engine is not None
            else None
        )

    def get_redis(self) -> Generator[Redis | None, None, None]:
        r = None
        try:
            r = self.__db.get_redis(force_new_instance=True)
            yield r
        finally:
            self.__db.try_close(r)

    def get_aredis(self) -> Generator[AsyncRedis, None, None]:
        """Get an async Redis client."""
        r = None
        r = self.__db.get_aredis(new_instance=True)
        yield r  # no need to close redis, it will manage itself

    def get_extended_redis(
        self, *, namespace: Optional[str] = None
    ) -> Generator[ExtendedRedis | None, None, None]:
        r = None
        try:
            r = ExtendedRedis.extend(redis=self.__db.get_redis(), namespace=namespace)
            yield r
        finally:
            pass

    def get_mongo(self) -> Generator[MongoClient | None, None, None]:
        m = None
        try:
            m = self.__db.get_mongo(new_instance=True)
            yield m
        finally:
            self.__db.try_close(m)

    def get_amongo(self) -> Generator[AsyncIOMotorClient, None, None]:
        """Get an async MongoDB client."""
        m = None
        try:
            m = self.__db.get_amongo()
            yield m
        finally:
            self.__db.try_close(m)

    def get_write_sql(self) -> Generator[Session | None, None, None]:
        s = None
        try:
            s = self.__db.get_sql_write_client(new_instance=True)
            yield s
        finally:
            self.__db.try_close(s)

    def get_read_sql(self) -> Generator[Session | None, None, None]:
        """Get a SQL session for reading operations."""
        s = None
        try:
            s = self.__db.get_sql_read_client(new_instance=True)
            yield s
        finally:
            self.__db.try_close(s)

    async def get_write_asql(
        self,
    ) -> AsyncGenerator[SQLModelAsyncSession | None, None]:
        """Get an async SQL session for writing operations."""
        if self.self_asql_write_session_maker is None:
            raise ValueError("[Db] async_sql_write_engine is not set")

        # 2) Use an async-generator dependency to yield and then close
        async with self.self_asql_write_session_maker() as session:
            yield session
            # async with already calls session.close() for you

        # s = None
        # try:
        #     s = self.__db.get_async_sql_write_client(new_instance=True)
        #     yield s
        # finally:
        #     await self.__db.try_close_async(s)

    async def get_awrite_sql(
        self,
    ) -> AsyncGenerator[SQLModelAsyncSession | None, None]:
        async for session in self.get_write_asql():
            yield session

    async def get_async_write_sql(
        self,
    ) -> AsyncGenerator[SQLModelAsyncSession | None, None]:
        async for session in self.get_write_asql():
            yield session

    async def get_read_asql(
        self,
    ) -> AsyncGenerator[SQLModelAsyncSession | None, None]:
        """Get an async SQL session for reading operations."""
        if self.self_asql_read_session_maker is None:
            raise ValueError("[Db] async_sql_read_engine is not set")

        # 2) Use an async-generator dependency to yield and then close
        async with self.self_asql_read_session_maker() as session:
            yield session
            # async with already calls session.close() for you
        # s = None
        # try:
        #     s = self.__db.get_async_sql_read_client(new_instance=True)
        #     yield s
        # finally:
        #     await self.__db.try_close_async(s)

    async def get_aread_sql(
        self,
    ) -> AsyncGenerator[SQLModelAsyncSession | None, None]:
        async for session in self.get_read_asql():
            yield session

    async def get_async_read_sql(
        self,
    ) -> AsyncGenerator[SQLModelAsyncSession | None, None]:
        async for session in self.get_read_asql():
            yield session

    def get_qdrant(self) -> Generator[QdrantClient | None, None, None]:
        q = None
        try:
            q = self.__db.get_qdrant_client(new_instance=True)
            yield q
        finally:
            self.__db.try_close(q)

    def get_aqdrant(self) -> Generator[AsyncQdrantClient | None, None, None]:
        """Get an async Qdrant client."""
        q = None
        try:
            q = self.__db.get_aqdrant_client(new_instance=True)
            yield q
        finally:
            self.__db.try_close(q)

    def get_extended_aredis(
        self, *, namespace: Optional[str] = None
    ) -> Generator[ExtendedAsyncRedis | None, None, None]:
        """Get an extended async Redis client with namespace support."""
        r = None
        r = ExtendedAsyncRedis.extend(
            aredis=self.__db.get_aredis(new_instance=True), namespace=namespace
        )
        yield r

    def get_neo4j_session(self) -> Generator[Neo4jSession, None, None]:
        with self.__db.get_neo4j_session(new_instance=True) as session:
            yield session

    async def get_aneo4j(
        self,
    ) -> AsyncGenerator[Neo4jAsyncTransaction, None]:
        async with self.__db.get_async_neo4j_session(new_instance=True) as session:
            async with session.begin_transaction() as transaction:
                yield transaction

    async def get_async_neo4j_session(
        self,
    ) -> AsyncGenerator[Neo4jAsyncSession, None]:
        async with self.__db.get_async_neo4j_session(new_instance=True) as session:
            yield session



from keble_helpers import PydanticModelConfig
from motor.motor_asyncio import AsyncIOMotorClient
from neo4j import AsyncTransaction
from pydantic import BaseModel
from qdrant_client import AsyncQdrantClient
from redis.asyncio import Redis as AsyncRedis

from keble_db.wrapper import ExtendedAsyncRedis


class AgentDbDeps(BaseModel):
    model_config = PydanticModelConfig.default(arbitrary_types_allowed=True)

    amongo: AsyncIOMotorClient
    aredis: AsyncRedis
    extended_aredis: ExtendedAsyncRedis
    aneo4j: AsyncTransaction
    aqdrant: AsyncQdrantClient



from typing import List, Optional, Tuple, Union, cast

from bson import ObjectId
from keble_helpers import is_pydantic_field_empty
from pymongo import ASCENDING, DESCENDING

from ..schemas import QueryBase, Uuid


def cleanse_query_for_first(query: Optional[QueryBase | dict]) -> dict:
    if query is None:
        return {}
    if isinstance(query, dict):
        query = QueryBase(**query)
    assert query.filters is None or isinstance(query.filters, dict), (
        "[Db] Query.filters only accept dict of Mongo query"
    )
    base = query.filters if query.filters is not None else {}
    base = __build_id_query(base, query)
    return base


def cleanse_query_for_list(query: Optional[QueryBase | dict]) -> dict:
    if query is None:
        return {}
    if isinstance(query, dict):
        query = QueryBase(**query)
    base = query.filters if query.filters is not None else {}
    assert query.filters is None or isinstance(query.filters, dict), (
        "[Db] Query.filters only accept dict of Mongo query"
    )
    assert query.offset is None or isinstance(query.offset, int), (
        "[Db] Offset must be int"
    )
    base = __build_ids_query(cast(dict, base), query)
    return base


def cleanse_query_order_by(
    query: Optional[QueryBase | dict],
) -> Optional[List[Tuple[str, int]]]:
    if isinstance(query, dict):
        query = QueryBase(**query)
    if query is not None and query.order_by is not None:
        assert isinstance(query.order_by, list), (
            "[Db] Mongodb accept only List[Tuple[str, ASCENDING | DESCENDING]] type ordering"
        )
        if len(query.order_by) == 0:
            return None
        for item in query.order_by:
            assert (
                isinstance(item, tuple)
                and len(item) == 2
                and isinstance(item[0], str)
                and item[1] in [DESCENDING, ASCENDING]
            ), (
                "[Db] Mongodb accept only List[Tuple[str, ASCENDING | DESCENDING]] type ordering"
            )
        return query.order_by


MaybeObjectId = Union[str, ObjectId]


def build_object_id(
    _id: Union[
        MaybeObjectId,
        List[MaybeObjectId],
        List[ObjectId | Uuid | str | int],
        Uuid,
        str | int | None,
    ],
) -> Union[ObjectId, List[ObjectId]]:
    if _id is None:
        return []
    if isinstance(_id, Uuid):
        raise ValueError("[Db] Uuid is not ObjectId")
    if isinstance(_id, int):
        raise ValueError("[Db] Int is not ObjectId")
    if isinstance(_id, list):
        return cast(List[ObjectId], [build_object_id(i) for i in _id])
    if isinstance(_id, str):
        return ObjectId(_id)
    return _id


def __build_id_query(base_query: dict, query: QueryBase) -> dict:
    if is_pydantic_field_empty(query, "id"):
        return base_query
    # use $and if necessary
    q = {"_id": build_object_id(query.id)}
    return {"$and": [base_query, q]} if "_id" in base_query else {**base_query, **q}


def __build_ids_query(base_query: dict, query: QueryBase) -> dict:
    if is_pydantic_field_empty(query, "ids"):
        return base_query
    # use $and if necessary
    q = {"_id": {"$in": build_object_id(query.ids)}}
    return {"$and": [base_query, q]} if "_id" in base_query else {**base_query, **q}


def ensure_object_id(id: MaybeObjectId) -> ObjectId:
    if isinstance(id, str):
        return ObjectId(id)
    return id



import json
from typing import Any, Awaitable, Callable, Generic, Optional, ParamSpec, TypeVar, cast

from keble_helpers import jsonable_encoder
from pydantic import BaseModel
from redis import Redis
from redis.asyncio import Redis as AsyncRedis


class ExtendedRedis(Redis):
    def __init__(self, *args, **kwargs):
        self._namespace: Optional[str] = None
        super().__init__(*args, **kwargs)

    @classmethod
    def extend(cls, redis: Redis, *, namespace: str | None = None):
        return ExtendedRedis._create_with_namespace(redis=redis, namespace=namespace)

    def clone_with_namespace(self, namespace: str) -> "ExtendedRedis":
        """
        Creates a new Redis instance with the same connection parameters.
        """
        return self._create_with_namespace(namespace=namespace, redis=self)

    def delete_namespace_keys(self):
        if self._namespace is None:
            raise ValueError("[Db] Namespace not set")
        # Pattern for keys to be deleted
        pattern = f"{self._namespace}:*"

        # Get keys matching the pattern
        keys = self.keys(pattern)
        # Delete the keys
        if keys:
            self.delete(*keys)  # type: ignore

    @classmethod
    def _create_with_namespace(
        cls, redis: Redis, *, namespace: str | None
    ) -> "ExtendedRedis":
        """
        Creates a new Redis instance with the same connection parameters.
        """
        # Retrieve the connection parameters from the current Redis instance
        connection_params = {
            "host": redis.connection_pool.connection_kwargs.get("host", "localhost"),
            "port": redis.connection_pool.connection_kwargs.get("port", 6379),
            "db": redis.connection_pool.connection_kwargs.get("db", 0),
            "password": redis.connection_pool.connection_kwargs.get("password", None),
            "socket_timeout": redis.connection_pool.connection_kwargs.get(
                "socket_timeout", None
            ),
            # Always ensure responses are decoded to strings
            "decode_responses": True,
        }

        # Create and return a new instance with the same connection params
        clone = ExtendedRedis(**connection_params)
        clone.set_namespace(namespace)
        return clone

    def set_namespace(self, namespace: str | None):
        self._namespace = namespace  # set namespace

    def get_namespace_key(self, key: str) -> str:
        if self._namespace is None:
            raise ValueError("[Db] Namespace not set")
        return f"{self._namespace}:{key}"


P = ParamSpec("P")
GetterReturnT = TypeVar("GetterReturnT")


class CacheOrSetRedisConfig(BaseModel, Generic[P, GetterReturnT]):
    ex: int = 3600  # Expiration time in seconds
    key: str
    getter_func: Callable[P, Awaitable[GetterReturnT]]
    serializer_func: Callable[[Any], GetterReturnT] | None = None


class ExtendedAsyncRedis(AsyncRedis):
    def __init__(self, *args, **kwargs):
        self._namespace: Optional[str] = None
        super().__init__(*args, **kwargs)

    @classmethod
    def extend(cls, aredis: AsyncRedis, *, namespace: str | None = None):
        return ExtendedAsyncRedis._create_with_namespace(
            aredis=aredis, namespace=namespace
        )

    def clone_with_namespace(self, namespace: str) -> "ExtendedAsyncRedis":
        """
        Creates a new Redis instance with the same connection parameters.
        """
        return self._create_with_namespace(namespace=namespace, aredis=self)

    def delete_namespace_keys(self):
        if self._namespace is None:
            raise ValueError("[Db] Namespace not set")
        # Pattern for keys to be deleted
        pattern = f"{self._namespace}:*"

        # Get keys matching the pattern
        keys = self.keys(pattern)
        # Delete the keys
        if keys:
            self.delete(*keys)  # type: ignore

    @classmethod
    def _create_with_namespace(
        cls, aredis: AsyncRedis, *, namespace: str | None
    ) -> "ExtendedAsyncRedis":
        """
        Creates a new Redis instance with the same connection parameters.
        """
        # Retrieve the connection parameters from the current Redis instance
        connection_params = {
            "host": aredis.connection_pool.connection_kwargs.get("host", "localhost"),
            "port": aredis.connection_pool.connection_kwargs.get("port", 6379),
            "db": aredis.connection_pool.connection_kwargs.get("db", 0),
            "password": aredis.connection_pool.connection_kwargs.get("password", None),
            "socket_timeout": aredis.connection_pool.connection_kwargs.get(
                "socket_timeout", None
            ),
            # Always ensure responses are decoded to strings
            "decode_responses": True,
        }

        # Create and return a new instance with the same connection params
        clone = ExtendedAsyncRedis(**connection_params)
        clone.set_namespace(namespace)
        return clone

    def set_namespace(self, namespace: str | None):
        self._namespace = namespace  # set namespace

    def get_namespace_key(self, key: str) -> str:
        if self._namespace is None:
            raise ValueError("[Db] Namespace not set")
        return f"{self._namespace}:{key}"

    async def acache_or_get(
        self,
        config: CacheOrSetRedisConfig[P, GetterReturnT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> GetterReturnT:
        cached_value = await self.get(config.key)
        if cached_value is not None:
            cached_value = json.loads(cached_value)
            if config.serializer_func is not None:
                return config.serializer_func(cached_value)
            return cast(GetterReturnT, cached_value)
        # If not in cache, call the getter function to retrieve the value
        value = await config.getter_func(*args, **kwargs)
        # Store the value in cache with an expiration time
        if value is not None:
            _json = jsonable_encoder(value)
            expire_seconds = config.ex
            await self.set(
                config.key,
                json.dumps(_json),
                ex=expire_seconds,
            )
        return value



import inspect
import logging
from typing import Dict, List, Literal, Mapping, Optional, Union, overload

from motor.motor_asyncio import AsyncIOMotorClient
from neo4j import AsyncDriver, AsyncGraphDatabase, Driver, GraphDatabase
from neo4j import AsyncSession as Neo4jAsyncSession
from neo4j import AsyncTransaction as Neo4jAsyncTransaction
from neo4j import Session as Neo4jSession
from pymongo import IndexModel, MongoClient
from qdrant_client import AsyncQdrantClient, QdrantClient
from qdrant_client.http import models as qdrant_models
from qdrant_client.http.exceptions import UnexpectedResponse
from qdrant_client.http.models import PayloadSchemaType
from redis import ConnectionPool
from redis import Redis as R
from redis.asyncio import ConnectionPool as AsyncConnectionPool
from redis.asyncio import Redis as AsyncRedis
from sqlalchemy import Engine
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine
from sqlmodel import Session, create_engine
from sqlmodel.ext.asyncio.session import AsyncSession as SQLModelAsyncSession

from keble_db.wrapper import ExtendedAsyncRedis
from keble_db.wrapper.redis import ExtendedRedis

from .schemas import DbSettingsABC

logger = logging.getLogger(__name__)


class Db:
    def __init__(self, settings: DbSettingsABC):
        self.__settings = settings
        self.__redis_pool = (
            ConnectionPool.from_url(settings.redis_uri, decode_responses=True)
            if settings.redis_uri is not None
            else None
        )
        self.__aredis_pool = (
            AsyncConnectionPool.from_url(settings.redis_uri, decode_responses=True)
            if settings.redis_uri is not None
            else None
        )
        self.__sql_read_engine = (
            create_engine(
                str(
                    settings.sql_read_uri
                    if settings.sql_read_uri is not None
                    else settings.sql_uri
                ),
                pool_pre_ping=True,
                pool_size=1000,
                max_overflow=10,
                pool_recycle=3600,
                pool_use_lifo=True,
                echo=False,
            )
            if settings.sql_read_uri is not None or settings.sql_uri is not None
            else None
        )
        self.__sql_write_engine = (
            create_engine(
                str(
                    settings.sql_write_uri
                    if settings.sql_write_uri is not None
                    else settings.sql_uri
                ),
                pool_pre_ping=True,
                pool_size=1000,
                max_overflow=10,
                pool_recycle=3600,
                pool_use_lifo=True,
                echo=False,
            )
            if settings.sql_write_uri is not None or settings.sql_uri is not None
            else None
        )

        # New async engines
        self.__async_sql_read_engine = (
            create_async_engine(
                str(
                    settings.sql_read_uri
                    if settings.sql_read_uri is not None
                    else settings.sql_uri
                ),
                pool_pre_ping=True,
                pool_size=100,
                max_overflow=10,
                pool_recycle=3600,
                echo=False,
            )
            if settings.sql_read_uri is not None or settings.sql_uri is not None
            else None
        )
        self.__async_sql_write_engine = (
            create_async_engine(
                str(
                    settings.sql_write_uri
                    if settings.sql_write_uri is not None
                    else settings.sql_uri
                ),
                pool_pre_ping=True,
                pool_size=100,
                max_overflow=10,
                pool_recycle=3600,
                echo=False,
            )
            if settings.sql_write_uri is not None or settings.sql_uri is not None
            else None
        )

        self._sql_read_instance = None
        self._sql_write_instance = None
        self._async_sql_read_instance = None
        self._async_sql_write_instance = None
        self._redis_instance = None
        self._aredis_instance = None
        self._mongo_instance = None
        self._amongo_instance: Optional[AsyncIOMotorClient] = None
        self._aqdrant_instance: Optional[AsyncQdrantClient] = None
        self._qdrant_instance = None
        self._neo4j_driver: Optional[Driver] = None
        self._neo4j_async_driver: Optional[AsyncDriver] = None
        self._neo4j_session: Optional[Neo4jSession] = None
        self._neo4j_async_session: Optional[Neo4jAsyncSession] = None

    @property
    def sql_write_engine(self) -> Optional[Engine]:
        return self.__sql_write_engine

    @property
    def sql_read_engine(self) -> Optional[Engine]:
        return self.__sql_read_engine

    @property
    def async_sql_write_engine(self) -> Optional[AsyncEngine]:
        return self.__async_sql_write_engine

    @property
    def async_sql_read_engine(self) -> Optional[AsyncEngine]:
        return self.__async_sql_read_engine

    def get_redis(self, *, force_new_instance: bool = False) -> R:
        """Return conn to redis
        You do not need to explicitly close the redis conn"""
        if self.__redis_pool is None:
            raise ValueError("[Db] redis_pool required")
        if force_new_instance:
            return R(connection_pool=self.__redis_pool, decode_responses=True)
        if not force_new_instance and self._redis_instance is not None:
            return self._redis_instance
        self._redis_instance = R(
            connection_pool=self.__redis_pool, decode_responses=True
        )
        return self._redis_instance

    def get_aredis(self, *, new_instance: bool = True) -> AsyncRedis:
        if self.__aredis_pool is None:
            raise ValueError("[Db] aredis_pool required")
        if new_instance:
            return AsyncRedis(connection_pool=self.__aredis_pool, decode_responses=True)
        if not new_instance and self._aredis_instance is not None:
            return self._aredis_instance
        self._aredis_instance = AsyncRedis(
            connection_pool=self.__aredis_pool, decode_responses=True
        )
        return self._aredis_instance

    def get_extended_aredis(
        self, *, new_instance: bool = True, namespace: Optional[str] = None
    ) -> ExtendedAsyncRedis:
        return ExtendedAsyncRedis.extend(
            aredis=self.get_aredis(new_instance=new_instance), namespace=namespace
        )

    def get_extended_redis(
        self, *, force_new_instance: bool = False, namespace: Optional[str] = None
    ) -> ExtendedRedis:
        return ExtendedRedis.extend(
            redis=self.get_redis(force_new_instance=force_new_instance),
            namespace=namespace,
        )

    def get_mongo(self, *, new_instance: bool = True) -> MongoClient:
        """Return a mongo conn
        You do not need to close PyMongo connections. Leave them open so that PyMongo connection pooling gives you the most efficient performance"""
        if self.__settings.mongo_db_uri is None:
            raise ValueError("[Db] mongo_db_uri required")

        if new_instance:
            return MongoClient(
                self.__settings.mongo_db_uri, uuidRepresentation="standard"
            )

        if self._mongo_instance is not None:
            try:
                self._mongo_instance.server_info()
                return self._mongo_instance
            except Exception as e:
                logger.critical(
                    f"[Db] Old mongo instance has been closed, now create a new one and returned. Server info exception: {e}"
                )
                pass

        self._mongo_instance = MongoClient(
            self.__settings.mongo_db_uri, uuidRepresentation="standard"
        )
        return self._mongo_instance

    async def aget_amongo(self, *, new_instance: bool = True) -> AsyncIOMotorClient:
        """Return a async mongo conn
        You do not need to close PyMongo connections. Leave them open so that PyMongo connection pooling gives you the most efficient performance"""
        if self.__settings.mongo_db_uri is None:
            raise ValueError("[Db] mongo_db_uri required")

        if new_instance:
            # check new instance first, new instance is independent and will not save as self.instance
            return AsyncIOMotorClient(
                self.__settings.mongo_db_uri, uuidRepresentation="standard"
            )

        if self._amongo_instance is not None:
            try:
                await self._amongo_instance.server_info()
                return self._amongo_instance
            except Exception as e:
                logger.critical(
                    f"[Db] Old async mongo instance has been closed, now create a new one and returned. Server info exception: {e}"
                )
                pass

        self._amongo_instance = AsyncIOMotorClient(
            self.__settings.mongo_db_uri, uuidRepresentation="standard"
        )
        return self._amongo_instance

    def get_amongo(self) -> AsyncIOMotorClient:
        """Return a async mongo conn
        without checking old instance, use with caustious to prevent creating too many session/conn to mongo at the same time.
        """
        if self.__settings.mongo_db_uri is None:
            raise ValueError("[Db] mongo_db_uri required")

        return AsyncIOMotorClient(
            self.__settings.mongo_db_uri, uuidRepresentation="standard"
        )

    def get_qdrant_client(self, *, new_instance: bool = True) -> QdrantClient:
        if self.__settings.qdrant_host is None or self.__settings.qdrant_port is None:
            raise ValueError("[Db] qdrant_host and qdrant_port required")
        if new_instance:
            # check new instance first, new instance is independent and will not save as self.instance
            return QdrantClient(
                host=self.__settings.qdrant_host,
                port=self.__settings.qdrant_port,
                timeout=120,
            )
        if not new_instance and self._qdrant_instance is not None:
            return self._qdrant_instance
        self._qdrant_instance = QdrantClient(
            host=self.__settings.qdrant_host,
            port=self.__settings.qdrant_port,
            timeout=120,
        )
        return self._qdrant_instance

    def get_aqdrant_client(self, *, new_instance: bool = True) -> AsyncQdrantClient:
        if self.__settings.qdrant_host is None or self.__settings.qdrant_port is None:
            raise ValueError("[Db] qdrant_host and qdrant_port required")
        if new_instance:
            # check new instance first, new instance is independent and will not save as self.instance
            return AsyncQdrantClient(
                host=self.__settings.qdrant_host,
                port=self.__settings.qdrant_port,
                timeout=120,
            )
        if not new_instance and self._aqdrant_instance is not None:
            return self._aqdrant_instance
        self._aqdrant_instance = AsyncQdrantClient(
            host=self.__settings.qdrant_host,
            port=self.__settings.qdrant_port,
            timeout=120,
        )
        return self._aqdrant_instance

    def get_neo4j_driver(self, *, new_instance: bool = True) -> Driver:
        if self.__settings.neo4j_uri is None:
            raise ValueError("[Db] neo4j_uri required")
        if self._neo4j_driver is None or new_instance:
            auth = None
            if (
                self.__settings.neo4j_user is not None
                and self.__settings.neo4j_password is not None
            ):
                auth = (self.__settings.neo4j_user, self.__settings.neo4j_password)
            self._neo4j_driver = GraphDatabase.driver(
                self.__settings.neo4j_uri,
                auth=auth,
            )
        return self._neo4j_driver

    def get_aneo4j_driver(self, *, new_instance: bool = True) -> AsyncDriver:
        if self.__settings.neo4j_uri is None:
            raise ValueError("[Db] neo4j_uri required")
        if self._neo4j_async_driver is None or new_instance:
            auth = None
            if (
                self.__settings.neo4j_user is not None
                and self.__settings.neo4j_password is not None
            ):
                auth = (self.__settings.neo4j_user, self.__settings.neo4j_password)
            self._neo4j_async_driver = AsyncGraphDatabase.driver(
                self.__settings.neo4j_uri,
                auth=auth,
            )
        return self._neo4j_async_driver

    def get_async_neo4j_driver(self, *, new_instance: bool = True) -> AsyncDriver:
        return self.get_aneo4j_driver(new_instance=new_instance)

    def get_neo4j_session(
        self,
        *,
        new_instance: bool = True,
        database: Optional[str] = None,
    ) -> Neo4jSession:
        driver = self.get_neo4j_driver(new_instance=new_instance)
        target_db = database if database is not None else self.__settings.neo4j_database
        if new_instance or self._neo4j_session is None:
            self._neo4j_session = driver.session(database=target_db)
        return self._neo4j_session

    @overload
    async def get_aneo4j(
        self,
        *,
        new_instance: bool = True,
        database: Optional[str] = None,
        as_transaction: Literal[True] = True,
    ) -> Neo4jAsyncTransaction: ...

    @overload
    async def get_aneo4j(
        self,
        *,
        new_instance: bool = True,
        database: Optional[str] = None,
        as_transaction: Literal[False],
    ) -> Neo4jAsyncSession: ...

    async def get_aneo4j(
        self,
        *,
        new_instance: bool = True,
        database: Optional[str] = None,
        as_transaction: bool = True,
    ) -> Neo4jAsyncTransaction | Neo4jAsyncSession:
        session = self.get_async_neo4j_session(
            new_instance=new_instance,
            database=database,
        )
        if not as_transaction:
            return session
        return await session.begin_transaction()

    def get_async_neo4j_session(
        self,
        *,
        new_instance: bool = True,
        database: Optional[str] = None,
    ) -> Neo4jAsyncSession:
        driver = self.get_aneo4j_driver(new_instance=new_instance)
        target_db = database if database is not None else self.__settings.neo4j_database
        if new_instance or self._neo4j_async_session is None:
            self._neo4j_async_session = driver.session(database=target_db)
        return self._neo4j_async_session

    def get_sql_write_client(self, *, new_instance: bool = True) -> Session:
        if self.sql_write_engine is None:
            raise ValueError("[Db] sql_write_engine required")
        if new_instance:
            return Session(self.sql_write_engine)
        if not new_instance and self._sql_write_instance is not None:
            return self._sql_write_instance
        self._sql_write_instance = Session(self.sql_write_engine)
        return self._sql_write_instance

    def get_sql_read_client(self, *, new_instance: bool = True) -> Session:
        if self.sql_read_engine is None:
            raise ValueError("[Db] sql_read_engine required")
        if new_instance:
            return Session(self.sql_read_engine)
        if not new_instance and self._sql_read_instance is not None:
            return self._sql_read_instance
        self._sql_read_instance = Session(self.sql_read_engine)
        return self._sql_read_instance

    # def get_async_sql_write_client(
    #     self, *, new_instance: bool = True
    # ) -> SQLModelAsyncSession:
    #     """Get an async SQL session for writing operations."""
    #     if self.async_sql_write_engine is None:
    #         raise ValueError("[Db] async_sql_write_engine required")
    #     if new_instance:
    #         return SQLModelAsyncSession(self.async_sql_write_engine, expire_on_commit=False)
    #     if not new_instance and self._async_sql_write_instance is not None:
    #         return self._async_sql_write_instance
    #     self._async_sql_write_instance = SQLModelAsyncSession(
    #         self.async_sql_write_engine,
    #         expire_on_commit=False  # Prevent greenlet_spawn error on attribute access after commit
    #     )
    #     return self._async_sql_write_instance

    # def get_async_sql_read_client(
    #     self, *, new_instance: bool = True
    # ) -> SQLModelAsyncSession:
    #     """Get an async SQL session for reading operations."""
    #     if self.async_sql_read_engine is None:
    #         raise ValueError("[Db] async_sql_read_engine required")
    #     if new_instance:
    #         return SQLModelAsyncSession(self.async_sql_read_engine, expire_on_commit=False)
    #     if not new_instance and self._async_sql_read_instance is not None:
    #         return self._async_sql_read_instance
    #     self._async_sql_read_instance = SQLModelAsyncSession(
    #         self.async_sql_read_engine,
    #         expire_on_commit=False  # Prevent greenlet_spawn error on attribute access after commit
    #     )
    #     return self._async_sql_read_instance

    @staticmethod
    def ready_mongo(client: MongoClient):
        # check by list
        client.list_databases()

    @staticmethod
    def ready_mongo_indexes(
        client: MongoClient,
        db_name: str,
        collection_name: str,
        indexes: List[Dict[str, Union[int, str]]],
    ) -> None:
        """
        Ensure MongoDB indexes exist. Creates them if not.

        :param client: MongoClient instance
        :param db_name: Name of the MongoDB database
        :param collection_name: Name of the MongoDB collection
        :param indexes: List of dictionaries defining the index fields and their types.
                        - For normal ascending/descending indexes, use 1 (ascending) or -1 (descending).
                        - For text search indexes, use "text".
                        Example: [{"field1": 1, "field2": -1}, {"field3": "text"}]
        """
        collection = client[db_name][collection_name]
        existing_indexes = collection.index_information()

        # Create index models based on the input
        indexes_need_to_create = []
        for index in indexes:
            compound_index_key = [(field, order) for field, order in index.items()]

            # Check if the index already exists in the collection
            if not any(
                compound_index_key == idx.get("key")
                for idx in existing_indexes.values()
            ):
                index_model = IndexModel(compound_index_key)
                indexes_need_to_create.append(index_model)
        if len(indexes_need_to_create) > 0:
            collection.create_indexes(indexes_need_to_create)

    @staticmethod
    def ready_redis(redis: R):
        # check by get key
        redis.get("k")

    @staticmethod
    def ready_qdrant(
        client: QdrantClient,
        collection: str,
        vectors_config: Union[
            qdrant_models.VectorParams, Mapping[str, qdrant_models.VectorParams]
        ],
    ):
        # need to check database exist or not
        try:
            client.get_collection(collection_name=collection)
        except (AssertionError, UnexpectedResponse):
            # collection not found, or connection error. Try to create
            client.create_collection(
                collection_name=collection,
                vectors_config=vectors_config,
            )

    @staticmethod
    def ready_neo4j(driver: Driver):
        # Verify connectivity; raises if misconfigured
        driver.verify_connectivity()

    @staticmethod
    def ready_qdrant_indexes(
        client: QdrantClient,
        collection_name: str,
        indexes: List[Dict[str, PayloadSchemaType]],
    ) -> None:
        """
        Ensure Qdrant payload indexes exist. Creates them if not.

        :param client: QdrantClient instance
        :param collection_name: Name of the Qdrant collection
        :param indexes: List of dictionaries where keys are field names and values are their corresponding types (PayloadSchemaType).
                        Example: [{"field1": PayloadSchemaType.INTEGER}, {"field2": PayloadSchemaType.KEYWORD}]
        """
        try:
            collection_info = client.get_collection(collection_name)
            existing_indexes = collection_info.payload_schema or {}

            for index in indexes:
                for field_name, schema_type in index.items():
                    if field_name not in existing_indexes:
                        client.create_payload_index(
                            collection_name=collection_name,
                            field_name=field_name,
                            field_schema=qdrant_models.PayloadSchemaType(schema_type),
                        )
        except (AssertionError, UnexpectedResponse):
            logger.critical(
                f"[Db] Collection {collection_name} does not exist or failed to retrieve collection info."
            )

    @staticmethod
    def try_close(
        *sessions: Optional[
            MongoClient
            | R
            | Session
            | QdrantClient
            | AsyncIOMotorClient
            | AsyncQdrantClient
            | AsyncRedis
            | SQLModelAsyncSession
            | AsyncEngine
            | Driver
            | AsyncDriver
            | Neo4jSession
            | Neo4jAsyncSession
            | Neo4jAsyncTransaction
        ],
    ):
        import asyncio

        for session in sessions:
            if session is None:
                continue

            try:
                # Call close() method
                close_method = getattr(session, "close", None)
                if close_method is None:
                    continue

                # Check if it's an async method
                if inspect.iscoroutinefunction(close_method):
                    # Create a coroutine
                    coro = close_method()

                    # Try to run it or create a task based on context
                    try:
                        # Check if we're in an event loop
                        loop = asyncio.get_event_loop()
                        if loop.is_running():
                            # Create a task - we can't await here
                            # Type ignores for linter errors - these are valid operations
                            asyncio.create_task(coro)  # type: ignore
                        else:
                            # No running loop, create one and run
                            asyncio.run(coro)  # type: ignore
                    except RuntimeError:
                        # No event loop exists, create one
                        try:
                            asyncio.run(coro)  # type: ignore
                        except Exception as e:
                            logger.error(f"[Db] Failed to run async close: {e}")
                else:
                    # Regular synchronous close
                    result = close_method()

                    # Check if the result is awaitable
                    if inspect.isawaitable(result):
                        try:
                            loop = asyncio.get_event_loop()
                            if loop.is_running():
                                asyncio.create_task(result)  # type: ignore
                            else:
                                asyncio.run(result)  # type: ignore
                        except RuntimeError:
                            try:
                                asyncio.run(result)  # type: ignore
                            except Exception as e:
                                logger.error(f"[Db] Failed to run awaitable: {e}")

            except Exception as e:
                # Log errors but don't raise
                logger.error(f"[Db] Failed to close a db session due to {e}")

    @staticmethod
    async def try_close_async(
        *sessions: Optional[
            MongoClient
            | R
            | Session
            | QdrantClient
            | AsyncIOMotorClient
            | AsyncQdrantClient
            | AsyncRedis
            | SQLModelAsyncSession
            | AsyncEngine
            | Driver
            | AsyncDriver
            | Neo4jSession
            | Neo4jAsyncSession
            | Neo4jAsyncTransaction
        ],
    ):
        """Async version of try_close that can properly await async close methods."""
        for session in sessions:
            if session is None:
                continue

            try:
                close_method = getattr(session, "close", None)
                if close_method is None:
                    continue

                if inspect.iscoroutinefunction(close_method):
                    await close_method()
                else:
                    result = close_method()
                    if inspect.isawaitable(result):
                        await result

            except Exception as e:
                logger.error(f"[Db] Failed to close async session due to {e}")
